@@ -108,51 +108,6 @@ def reduce_whitespace(data):
     return data.apply(lambda x: x.str.strip().str.replace(" +", " ", regex=True))
 
 
-def replicate_address(data, i, street_suffix_mapping, unit_mapping):
-    """
-    Replicate street addresses 
-    
-    Parameters
-    ----------
-    data: pd.DataFrame
-       ACS dataFrame to make changes to 
-    street_address: str
-       Name of street address column 
-    street_suffix_mapping: dict
-       Dictionary with street mappings
-    unit_mapping: dict
-       Dictionary with building unit mapping   
-    """
-    df_base =  data[i]     # base is complete, containing the original record (1)
-
-    data[i] = pd.DataFrame([{i: data[i]}]).replace(unit_mapping, regex=True).loc[0, i]
-    df_u = data[i].strip()
-    data[i] = pd.DataFrame([{i: data[i]}]).replace(street_suffix_mapping, regex=True).loc[0, i] # Second Unit & Street Suffix abbreviation mapping
-    
-    df_sus =  data[i]     # street & unit mapping + split after suffix (2)
-    df_suu =  data[i]     # street & unit mapping + split before unit (3)
-    df_sus = df_sus.split("-", 1)[0].strip()
-    df_u = df_u.split("-", 1)[0]
-    df_suu = df_suu.split("-", 1)[0].strip()
-    data[i] = data[i].replace("-", "").strip()
- 
-    df_su =  data[i]      # suffix & unit mapped to abbreviations (4) 
-    df_no =  data[i]      # remove all numbers not-attached to str (5)
-
-    df_no = re.sub("^([0-9]{1,6}[A-Z]{1,3})", "", df_no)
-    df_no = re.sub(" [0-9]{4,7} ", "", df_no)
-
-    dataout = pd.DataFrame([df_base,
-                           df_u,
-                           df_su,
-                           df_sus,
-                           df_suu,
-                           df_no
-                          ])
-    dataout.index = [i, f"{i}_u", f"{i}_su", f"{i}_sus", f"{i}_suu",f"{i}_no" ]
-    dataout = dataout.drop_duplicates()
-    return(dataout)
-
 
 def address_mining(data, i):
     """
@@ -172,7 +127,166 @@ def address_mining(data, i):
                                           str(data[i])))
     return(data[i])
 
+def replicate_house_number(data, house_number, add_to_flg):
+    """
+    Replicate street addresses 
+    
+    Parameters
+    ----------
+    data: pd.DataFrame
+        DataFrame to make changes to 
+    house_number: str
+        Name of street address column
+    add_to_flg: int
+        'replicate_flg' column indicates the order of preference of replicated rows. The smaller the flag the higher the preference. 'add_to_flg' value is added to the flag of replicated rows.
 
+    """
+    print("         ...Base")
+    df_base =  data.copy()# base is complete, containing the original record (1)
+    data['replicate_flg'] = data['replicate_flg'].apply(lambda s: str(int(s) + add_to_flg).zfill(6))
+    print("         ...Number processing...")
+    data[house_number].apply(lambda x: re.sub("[^0-9]","",str(x))) 
+    dataout = pd.concat([df_base, data], axis=0)
+    dataout = dataout.drop_duplicates(keep = 'first', subset = [col for col in dataout.columns if col != 'replicate_flg'])
+    print(f"         House number dataframe expansion is complete! (n={len(dataout)})")
+    return(dataout)
+    
+def replicate_address_2(data, street_address, street_suffix_mapping, add_to_flg = 0, replicate_with_flg = 1):
+    """
+    Replicate street addresses 
+    
+    Parameters
+    ----------
+    data: pd.DataFrame
+        DataFrame to make changes to 
+    street_address: str
+       Name of street address column 
+    street_suffix_mapping: dict
+       Dictionary with street mappings
+    add_to_flg: int
+        'replicate_flg' column indicates the order of preference of replicated rows. The smaller the flag the higher the preference. 'add_to_flg' value is added to the flag of replicated rows.
+    replicate_with_flg: bool
+        Flag indicating whether "replicate_flg" column needs to be edited
+    """
+
+    print("         ...Base")
+    df_base =  data.copy()
+    if replicate_with_flg ==1:
+        data['replicate_flg'] = data['replicate_flg'].apply(lambda s: str(int(s) + add_to_flg).zfill(6))
+    print("         ...Map street suffixes...")
+    data[street_address] = data[street_address].replace(street_suffix_mapping, regex=True) # this mapping takes the longest but is ok for 10K records
+    data[street_address] = np.where(data[street_address]=='nan', None, data[street_address])
+    print("         ...Mapped & split by street suffixes...")
+    # Remove addtl "-"
+    data[street_address] = data[street_address].str.split("-", 1, expand=True)[0]
+    data[street_address] = data[street_address].str.replace(pat = "-", repl = "", regex=False)
+    print("         ...Number processing...")
+    data[street_address] = data[street_address].str.replace("^([0-9]{1,6}[A-Z]{1,3})", "", regex=True)
+    data[street_address] = data[street_address].str.replace("^([0-9]{1,3} [A-Z]{2})", "", regex=True)
+    data[street_address] = data[street_address].str.replace(" [0-9]{4,7} ", "", regex=True)
+    print("")
+    dataout = pd.concat([df_base, data], axis=0)
+    dataout = dataout.drop_duplicates(keep = 'first', subset = [col for col in dataout.columns if col != 'replicate_flg'])
+    print(f"         Address dataframe expansion is complete! (n={len(dataout)})")
+    return(dataout)
+        
+
+def replicate_north_n(data, street_address, add_to_flg = 0, replicate_with_flg = 1):
+    """
+    Replicate street address by simplify cardinal directions. 
+    
+    Parameters
+    ----------
+    data: pd.DataFrame
+        DataFrame to make changes to 
+    street_address: str
+       Name of street address column 
+    add_to_flg: int
+        'replicate_flg' column indicates the order of preference of replicated rows. The smaller the flag the higher the preference. 'add_to_flg' value is added to the flag of replicated rows.
+    replicate_with_flg: bool
+        Flag indicating whether "replicate_flg" column needs to be edited  
+    """
+    north_n_mapping = { '\\bNORTH\\b' : 'N',
+                        '\\bSOUTH\\b' : 'S',
+                        '\\bEAST\\b' : 'E',
+                        '\\bWEST\\b' : 'W',
+                        '\\bNORTHEAST\\b' : 'NE',
+                        '\\bNORTHWEST\\b' : 'NW',
+                        '\\bSOUTHEAST\\b' : 'SE',
+                        '\\bSOUTHWEST\\b' : 'SW',
+                        '\\bNORTE\\b' : 'N',
+                        '\\bSUR\\b' : 'S' ,
+                        '\\bESTE\\b' : 'E',
+                        '\\bOESTE\\b' : 'O',
+                        '\\bNORESTE\\b' : 'NE',
+                        '\\bNOROESTE\\b' : 'NO',
+                        '\\bSUDESTE\\b' : 'SE',
+                        '\\bSUDOESTE\\b' : 'SO'}
+
+    df_base =  data.copy()
+    data[street_address] = data[street_address].replace(north_n_mapping, regex=True)
+    if replicate_with_flg == 1:
+        data['replicate_flg'] = data['replicate_flg'].apply(lambda s: str(int(s) + add_to_flg).zfill(6))
+    data[street_address] = data[street_address].replace(north_n_mapping, regex=True)
+    dataout = pd.concat([df_base, data], axis=0)
+    dataout = dataout.drop_duplicates(keep = 'first', subset = [col for col in dataout.columns if col != 'replicate_flg'])
+    return dataout
+       
+def split_char_position(a_string):
+    """
+    Returns position used to split a non-numeric house number 
+    
+    Parameters
+    ----------
+    a_string: str
+        String to be investigated 
+    """
+    position = [i for i, s in enumerate(a_string) if not s.isdigit()][-1] + 1
+    return position
+
+def split_HN(df, cols_to_split):
+    """
+    Returns DataFrame where non-numeric house numbers are split into two columns. One numeric and one non-numeric. 
+    
+    Parameters
+    ----------
+    df: pd.DataFrame
+        DataFrame to be modified
+    cols_to_split: list
+        List of columns containing house numbers
+    """
+    numeric_map = (df[cols_to_split[0]].str.isnumeric()) | (df[cols_to_split[0]].isna()) | (df[cols_to_split[0]] == '')
+    df_numeric = df[numeric_map]
+    df_non_numeric = df[~numeric_map]
+    
+    for col in cols_to_split:  
+        df_numeric[f'{col}_LEFT'] = ''
+        df_numeric[f'{col}_RIGHT'] = df_numeric[col]
+        
+        df_non_numeric[f'{col}_LEFT']  = df_non_numeric[col].apply(lambda a_string: a_string[:split_char_position(a_string)])
+        df_non_numeric[f'{col}_RIGHT'] = df_non_numeric[col].apply(lambda a_string: a_string[split_char_position(a_string):])
+    
+    return pd.concat([df_numeric, df_non_numeric])
+
+def sort_HN_columns(df):
+    """
+    Returns DataFrame where non-numeric house numbers are split into two columns. One numeric and one non-numeric. 
+    
+    Parameters
+    ----------
+    df: pd.DataFrame
+        DataFrame to be modified
+    cols_to_split: list
+        List of columns containing house numbers
+    """
+    df['FROMHN_RIGHT'] = pd.to_numeric(df['FROMHN_RIGHT'])
+    df['TOHN_RIGHT']   = pd.to_numeric(df['TOHN_RIGHT'])
+    from_hn = np.where(df['FROMHN_RIGHT'] <= df['TOHN_RIGHT'], df['FROMHN_RIGHT'], df['TOHN_RIGHT'])
+    to_hn   = np.where(df['FROMHN_RIGHT'] <= df['TOHN_RIGHT'], df['TOHN_RIGHT'], df['FROMHN_RIGHT'])
+    df['FROMHN_RIGHT'] = from_hn
+    df['TOHN_RIGHT'] = to_hn
+    return(df)
+    
 class HandleTracts():
     """
     HandleTracts parses the American Community Survey response format into state, county, and tracts.
@@ -203,74 +317,7 @@ class HandleTracts():
         acs_zip_split(acs_zip, feature="result")
         
         return(acs_ct, acs_zip)
-
     
-def replicate_address_2(data, street_address, street_suffix_mapping, unit_mapping):
-    """
-    Replicate street addresses 
-    
-    Parameters
-    ----------
-    data: pd.DataFrame
-        DataFrame to make changes to 
-    street_address: str
-       Name of street address column 
-    street_suffix_mapping: dict
-       Dictionary with street mappings
-    unit_mapping: dict
-       Dictionary with building unit mapping   
-    """
-    # base
-    data = data.reset_index(drop=False)
-    print("         ...Base")
-    df_base =  data.copy()# base is complete, containing the original record (1)
-    print("         ...Map street suffixes...")
-    data[street_address] = data[street_address].replace(street_suffix_mapping, regex=True) # this mapping takes the longest but is ok for 10K records
-    data[street_address] = np.where(data[street_address]=='nan', None, data[street_address])
-    print("         ...Mapped & split by street suffixes...")
-    # Remove addtl "-"
-    data[street_address] = data[street_address].str.split("-", 1, expand=True)[0]
-    data[street_address] = data[street_address].str.replace(pat = "-", repl = "", regex=False)
-    print("         ...Number processing...")
-    data[street_address] = data[street_address].str.replace("^([0-9]{1,6}[A-Z]{1,3})", "", regex=True)
-    data[street_address] = data[street_address].str.replace("^([0-9]{1,3} [A-Z]{2})", "", regex=True)
-    data[street_address] = data[street_address].str.replace(" [0-9]{4,7} ", "", regex=True)
-    print("")
-    dataout = pd.concat([df_base,
-                           data
-                          ], axis=0)
-    dataout = dataout.drop_duplicates()
-    print(f"         Address dataframe expansion is complete! (n={len(dataout)})")
-    return(dataout)
-
-
-def replicate_house_number(data, house_number):
-    """
-    Replicate street addresses 
-    
-    Parameters
-    ----------
-    data: pd.DataFrame
-        DataFrame to make changes to 
-    house_number: str
-       Name of street address column 
-    """
-    # base
-    data = data.reset_index(drop=True)
-    print("         ...Base")
-    df_base =  data.copy()# base is complete, containing the original record (1)
-    print("         ...Number processing...")
-    data[house_number].apply(lambda x: re.sub("[^0-9]",\
-                                                     "",\
-                                                    str(x))) 
-    dataout = pd.concat([df_base,
-                           data
-                          ], axis=0)
-    dataout = dataout.drop_duplicates()
-    print(f"         House number dataframe expansion is complete! (n={len(dataout)})")
-    return(dataout)
-
-       
 class ProcessStrings(BaseZRP):
     """
     ProcessStrings executes all ZRP preprocessing. All user data is processed with additional  processing operations for geo-specific and American Community Survey data.
@@ -279,25 +326,27 @@ class ProcessStrings(BaseZRP):
     ----------
     key: str 
         Key to set as index. If not provided, a key will be generated.
-    first_name: str
+    first_name: str 
         Name of first name column
-    middle_name: str
+    middle_name: str 
         Name of middle name column
-    last_name: str
+    last_name: str 
         Name of last name/surname column
-    house_number: str
+    house_number: str 
         Name of house number column. Also known as primary address number this is the unique number assigned to a building to delineate it from others on a street. This is usually the first component of a delivery address line.
-    street_address: str
+    street_address: str (required)
         Name of street address column. The street address is usually comprised of predirectional, street name, and street suffix. 
     city: str
         Name of city column
-    state: str
+    state: str 
         Name of state column
-    zip_code: str
+    zip_code: str 
         Name of zip or postal code column
-    census_tract: str
+    block_group: str (optional)
+        Name of block group column
+    census_tract: str (optional)
         Name of census tract column
-    support_files_path:
+    support_files_path: str (optional)
         File path with support data
     street_address_2: str, optional
         Name of additional address column
@@ -305,9 +354,9 @@ class ProcessStrings(BaseZRP):
         Name of column containing full name preix (ie Dr, Sr, and Esq )
     name_suffix: str, optional
         Name of column containing full name suffix (ie jr, iii, and phd)
-    na_values: list
+    na_values: list, optional
         List of missing values to replace 
-    file_path: str
+    file_path: str, optional
         Input data file path
     geocode: bool
         Whether to geocode
@@ -327,17 +376,20 @@ class ProcessStrings(BaseZRP):
     def fit(self, data):
         data_cols = list(data.columns)
         print("   [Start] Validating input data")
+        base_req = [self.first_name, self.middle_name, self.last_name, self.state, self.zip_code, self.street_address]
         if self.census_tract in data_cols:
-            self.required_cols = [self.first_name, self.middle_name, self.last_name, self.census_tract]
-        elif (self.census_tract in data_cols) & (self.block_group in data_cols):
-            self.required_cols = [self.first_name, self.middle_name, self.last_name, self.census_tract, self.block_group]
-        elif (self.zip_code in data_cols) & (self.geocode==True):
-            self.required_cols = [self.first_name, self.middle_name, self.last_name, self.zip_code, self.house_number, self.street_address, self.city, self.state]
-        elif (self.zip_code in data_cols) & (self.geocode==False):
-            self.required_cols = [self.first_name, self.middle_name, self.last_name, self.zip_code]
+            self.required_cols = base_req + [self.census_tract]
+        if self.block_group in data_cols:
+            self.required_cols = base_req + [self.block_group]
+        if (self.census_tract in data_cols) & (self.block_group in data_cols):
+            self.required_cols = base_req + [self.census_tract, self.block_group]
+        else:
+            self.required_cols = base_req + [self.house_number]
+            
+            
         val_na = is_missing(data, self.required_cols)
         if val_na:
-            assert True, f"     Missing required data {val_na}"
+            raise ValueError(f"     Missing required data {val_na}")            
         validate = ValidateInput()
         validate.fit()
         validators_in = validate.transform(data)
@@ -369,11 +421,6 @@ class ProcessStrings(BaseZRP):
         data_cols =  data.columns
         data = set_id(data, self.key)
         
-        numeric_cols =  list(set([self.zip_code,
-                                  self.census_tract,
-                                  self.house_number
-                                 ]).intersection(set(data_cols)))
-
         # Convert to uppercase & trim whitespace
         print("   Formatting P1")
         data = data.astype(str)
@@ -384,6 +431,12 @@ class ProcessStrings(BaseZRP):
                     "^NAN$": None,
                     "^NONE$": None}
         data = data.replace(na_dict, regex=True)     
+        data[self.house_number] = data[self.house_number].apply(lambda x: re.sub("[^0-9]$", "", str(x)))
+        data = split_HN(data, [self.house_number])
+        
+        numeric_cols =  list(set([self.zip_code,
+                                  self.census_tract,
+                                 ]).intersection(set(data_cols)))
         
         # Remove/replace special characters
         for col in numeric_cols:
@@ -449,7 +502,7 @@ class  ProcessACS(BaseZRP):
     
 class  ProcessGeo(BaseZRP):
     """
-    All user data is processed with additional  processing operations for geo-specific and American Community Survey data.
+    All user data is processed with additional processing operations for geo-specific and American Community Survey data.
 
     Parameters
     ----------
@@ -512,7 +565,7 @@ class  ProcessGeo(BaseZRP):
         val_na = is_missing(data, self.required_cols)
         val_na = is_missing(data, self.required_cols)
         if val_na:
-            assert True, f"Missing required data {val_na}"
+            raise ValueError(f"Missing required data {val_na}")
         geo_validate = ValidateGeo()
         geo_validate.fit()
         geo_validators_in = geo_validate.transform(data)
@@ -574,8 +627,7 @@ class  ProcessGeo(BaseZRP):
                                                                          str(x))))
         
         data_path = join(curpath, f'../data/processed')
-        state_mapping, street_suffix_mapping, directionals_mapping, unit_mapping = load_mappings(data_path)
-        
+                
         if self.key in data.columns:
             data["ZEST_KEY_COL"] = data[self.key]
         else:
@@ -586,8 +638,8 @@ class  ProcessGeo(BaseZRP):
         street_addr_results = Parallel(n_jobs = self.n_jobs, prefer="threads", verbose=1)(delayed((address_mining))(street_addr_dict, i) for i in tqdm(list(data.index)))
 
         data[self.street_address] = street_addr_results
-#         data[self.city]  = data[self.city].str.replace("[^\\w\\s]", "", regex=True)
-
+        
+        state_mapping, street_suffix_mapping, directionals_mapping, unit_mapping = load_mappings(data_path)
         # State
         data[self.state]  = data[self.state].str.replace("[^\\w\\s]", "", regex=True)
         data[self.state] = data[self.state].replace(state_mapping)
@@ -596,25 +648,29 @@ class  ProcessGeo(BaseZRP):
                                  (data[self.zip_code].str.contains("None")),
                                   None,
                                   data[self.zip_code].apply(lambda x: x.zfill(5)))
-        data[self.zip_code] = data[self.zip_code].astype(str).str[:5]        
-        
+        data[self.zip_code] = data[self.zip_code].astype(str).str[:5]
+        data['replicate_flg'] = '000000'
         if replicate:
             print("      ...replicating address")
-            data = replicate_address_2(data, self.street_address, street_suffix_mapping, unit_mapping)
-            data = replicate_house_number(data, self.house_number)
-        
+            data = replicate_house_number(data, self.house_number, 1)
+            data = replicate_address_2(data, self.street_address, street_suffix_mapping, 10, 1)
+            data = replicate_north_n(data, self.street_address, 100, 1)
+#             data = replicate_address_3(data, self.street_address, street_suffix_mapping_new_only, 1000)
+#             data = replicate_n_north(data, self.street_address, 1000)
+#             data = replicate_n_norte(data, self.street_address, 2000)
+        data = data.reset_index(drop=False)
         print("      ...formatting")
         addr_cols = list(set(list(data.columns)).intersection(set([self.zip_code, self.census_tract, self.house_number, self.city, self.state, self.street_address])))
 
 
         data = reduce_whitespace(data)
         print("   [Completed] Processing geo data")
-        return(data)
+        return data
 
     
 class  ProcessGLookUp(BaseZRP):
     """
-    All user data is processed with additional  processing operations for geo-specific and American Community Survey data.
+    All user data is processed with additional processing operations for geo-specific and American Community Survey data.
     Parameters
     ----------
     key: str 
@@ -726,11 +782,18 @@ class  ProcessGLookUp(BaseZRP):
         spec_cols = list(set(list(data_cols)).intersection(set([self.zip_code, self.block_group, self.county, self.state, self.census_tract])))
         data[spec_cols] = norm_na(data[spec_cols], self.na_values)
         data = reduce_whitespace(data)
+        data = split_HN(data, ['FROMHN', 'TOHN'])
+        data = sort_HN_columns(data)
+        
+        curpath = dirname(__file__)
+        data_path = join(curpath, '../data/processed')
+        _, street_suffix_mapping, _, _ = load_mappings(data_path)
+        data = replicate_address_2(data, self.street_address, street_suffix_mapping, replicate_with_flg = 0)
+        data = replicate_north_n(data, self.street_address, replicate_with_flg = 0)
+        
         print("   [Completed] Processing lookup data")
         
-        data_cols = list(data.columns)
         validate = ValidateInput()
         validate.fit()
         validator_in = validate.transform(data) 
-        print(validator_in)
         return(data)
